
import Base32._

import Calate._

/** GeoHash encoding/decoding as per http://en.wikipedia.org/wiki/Geohash */
object GeoHash {

  val LAT_RANGE = (-90.0, 90.0)
  val LON_RANGE = (-180.0, 180.0)

  // Aliases, utility functions
  type Bounds = (Double, Double)
  private def mid(b: Bounds) = (b._1 + b._2) / 2.0
  implicit class BoundedNum(x: Double) { def in(b: Bounds): Boolean = x >= b._1 && x <= b._2 }

  /**
    * Encode lat/long as a base32 geohash.
    *
    * Precision (optional) is the number of base32 chars desired; default is 12, which gives precision well under a meter.
    */
  def encode(lat: Double, lon: Double, precision: Int=12): String = { // scalastyle:ignore
    require(lat in LAT_RANGE, "Latitude out of range")
    require(lon in LON_RANGE, "Longitude out of range")
    require(precision > 0, "Precision must be a positive integer")
    val rem = precision % 2 // if precision is odd, we need an extra bit so the total bits divide by 5
    val numbits = (precision * 5) / 2
    val latBits = findBits(lat, LAT_RANGE, numbits)
    val lonBits = findBits(lon, LON_RANGE, numbits + rem)
    val bits = intercalate(lonBits, latBits) // interleave bits
    bits.grouped(5).map(toBase32).mkString // scalastyle:ignore
    // groups the bits 5 by 5 and converts each 5 bits into a base32 string which is the hash
  }

  private def findBits(part: Double, bounds: Bounds, p: Int): List[Boolean] = {
    if (p == 0) Nil // continue until the amount of bits is used up
    else {
      val avg = mid(bounds) // find the average value in the interval
      if (part >= avg) true :: findBits(part, (avg, bounds._2), p - 1) // >= to match geohash.org encoding
        // if the lat/lon is bigger than the avg, use the higher interval, recursive
      else false :: findBits(part, (bounds._1, avg), p - 1) // else use the lower interval
    }
  }

  /**
    * Decode a base32 geohash into a tuple of (lat, lon)
    */
  def decode(hash: String): (Double, Double) = {
    require(isValid(hash), "Not a valid Base32 number")
    val (odd, even) = extracalate(toBits(hash))
    val lon = mid(decodeBits(LON_RANGE, odd))
    val lat = mid(decodeBits(LAT_RANGE, even))
    (lat, lon)
  }

  private def decodeBits(bounds: Bounds, bits: Seq[Boolean]) =
    bits.foldLeft(bounds)((acc, bit) => if (bit) (mid(acc), acc._2) else (acc._1, mid(acc)))


  def adjacent(geohash : String, direction: Char) : String= {
    val BASESEQUENCE = "0123456789bcdefghjkmnpqrstuvwxyz"
    val neighbor = Map(
      'n' -> List("p0r21436x8zb9dcf5h7kjnmqesgutwvy", "bc01fg45238967deuvhjyznpkmstqrwx"),
      's' -> List("14365h7k9dcfesgujnmqp0r2twvyx8zb", "238967debc01fg45kmstqrwxuvhjyznp"),
      'e' -> List("bc01fg45238967deuvhjyznpkmstqrwx", "p0r21436x8zb9dcf5h7kjnmqesgutwvy"),
      'w' -> List("238967debc01fg45kmstqrwxuvhjyznp", "14365h7k9dcfesgujnmqp0r2twvyx8zb")
    )

    val border = Map(
      'n' -> List("prxz", "bcfguvyz"),
      's' -> List("028b", "0145hjnp"),
      'e' -> List("bcfguvyz", "prxz"),
      'w' -> List("0145hjnp", "028b")
    )

    val last = geohash.charAt(geohash.length-1)
    var parent = geohash.substring(0,geohash.length-1)
    val t = geohash.length % 2
    if (border(direction)(t).contains(last)){
      parent = adjacent(parent, direction)
    }
    parent + BASESEQUENCE.charAt(neighbor(direction)(t).indexOf(last))
  }

    def neighbors(geohash : String):Map[String,String] ={
       val map = Map(
         "n" -> adjacent(geohash, 'n'),
         "ne" -> adjacent(adjacent(geohash, 'n'), 'e'),
         "e" -> adjacent(geohash, 'e'),
         "se" -> adjacent(adjacent(geohash, 's'), 'e'),
         "s" -> adjacent(geohash, 's'),
         "sw" -> adjacent(adjacent(geohash, 's'), 'w'),
         "w" -> adjacent(geohash, 'w'),
         "nw" -> adjacent(adjacent(geohash, 'n'), 'w'),
         "c" -> geohash
       )
      map
    }

}

/**
  * This file is an implementation of Geohash made by davidallsopp and made public on github.com
  *
  * URL: https://github.com/davidallsopp/geohash-scala/tree/master/src/main/scala/com/github/davidallsopp
  *
  */

/**
  * Finding the neighbors algorithm was adapted from Python ref (https://github.com/transitland/mapzen-geohash/blob/master/mzgeohash/geohash.py)
  * into scala by me.
  *
  * */

