/**
  * Created by tord on 20.02.17.
  */
import org.apache.spark.sql._
import org.apache.spark.sql.functions.monotonically_increasing_id
import GeoHash._

import scala.util.Random
import com.fasterxml.jackson.databind.ObjectMapper
import org.apache.hadoop.fs.{FileSystem, Path}


case class ShortTweet(
                 tweetContent: String,
                 latitude: Double,
                 longitude: Double,
                 data : String
                )


object SpatioTextualIndex {

  def build(spark: SparkSession,
            inputPath: String,
            indexPrecision: Int = 12,
            queryPrecision: Int,
            indexPartitionPrecision: Int): Unit ={
    import spark.implicits._

    val stopwords = spark.read.textFile("stopwords")
    val stopwordsBroadcst = spark.sparkContext.broadcast(stopwords.collect().toSet)


    def geoHashAndFindWords2(zippedDataset : DataFrame) ={
      println("Geohashing zipped Dataset")
      val DFRemovedStopwords = zippedDataset.flatMap(row=> {
        val geohash = encode(row.getDouble(1),row.getDouble(2), indexPrecision)
        row.getString(0).split(" ")
          .map(_.replaceAll("[^A-Za-z0-9]", "").toLowerCase().trim)
          .filter(word => !word.contains("http") && !stopwordsBroadcst.value.contains(word))
          .map((_,geohash,geohash.charAt(0).toString, row.getLong(4)))
        }
      )
      stopwordsBroadcst.unpersist()
      DFRemovedStopwords
    }

    val out = spark.sqlContext.getConf("userOutputPath")

    val inputDataset = spark.read.textFile(inputPath)
      .filter(line => line.startsWith("{"))
      .map(line =>{
    val mapper = new ObjectMapper()
      try{
      val obj = mapper.readTree(line)
      if(obj.get("coordinates").isNull){
        ShortTweet(
          tweetContent = obj.get("text").asText("null"),
          longitude = obj.get("place").get("bounding_box").get("coordinates").get(0).get(0).get(0).asDouble(),
          latitude = obj.get("place").get("bounding_box").get("coordinates").get(0).get(0).get(1).asDouble(),
          data = ""
        )
      }else{
          ShortTweet(
            tweetContent = obj.get("text").asText("null"),
            longitude = obj.get("coordinates").get("coordinates").get(0).asDouble(0.0),
            latitude = obj.get("coordinates").get("coordinates").get(1).asDouble(0.0),
            data = ""
          )
        }
      }catch {
        case e: Exception => {
          println(e.getStackTrace)
          println(e.getMessage)
          e.printStackTrace()
          ShortTweet("",0.0,0.0,"")
        }
      }
    })
      .filter(obj => !obj.tweetContent.isEmpty)
      .withColumn("docId", monotonically_increasing_id())
      .withColumn("partition", $"docId" % 32)
      .write.partitionBy("partition").parquet(out + "zipped.parquet")

    val geo = geoHashAndFindWords2(spark.read.parquet(out + "zipped.parquet"))
      .write.partitionBy("_3").parquet(out + "geo.parquet")

  }

  def fullQuery(queryLocationInput: (Double, Double),
                queryTerm: String = scala.io.StdIn.readLine(),
                queryPrecision: Int,
                spark : SparkSession): Unit = {
    import spark.implicits._
    val out = spark.sqlContext.getConf("userOutputPath")
    val geoHash = neighbors(encode(queryLocationInput._1, queryLocationInput._2, queryPrecision)).values.toList
    val geohashBroadcast = spark.sparkContext.broadcast(geoHash)


    def performLocationQuery(datasetToQuery: DataFrame,
                             queryLocation: (Double, Double),
                             queryPrecision: Int):DataFrame = {

    val result = datasetToQuery
      .filter($"_2".substr(1,queryPrecision).isin(geohashBroadcast.value:_*))
    geohashBroadcast.unpersist()
    result
    }

    def performTermQuery(locationQueryResults: DataFrame,
                         queryTerm: String = scala.io.StdIn.readLine()): DataFrame= {
      val res = locationQueryResults.filter($"_1" equalTo queryTerm)
      res
    }

    val ghash = encode(queryLocationInput._1, queryLocationInput._2, queryPrecision)
    val geoHashedTweets = spark.read.parquet(out + "geo.parquet/_3="+ ghash.charAt(0).toString)



    val spatialResult= performLocationQuery(geoHashedTweets, queryLocationInput, queryPrecision)
    val spatialTextResult = performTermQuery(spatialResult, queryTerm)
      .withColumn("partition", $"_4" % 32).cache()
    val partitionCheckBroadcast = spark.sparkContext.broadcast(
    spatialTextResult.select("partition").distinct().as[Long].map(partition => out + "zipped.parquet/partition=" + partition).collect().toList
    )

    val zippedDataset = spark.read.parquet(partitionCheckBroadcast.value:_*)
    val results = zippedDataset.join(spatialTextResult.select("_4").distinct(), $"docId" === $"_4")
      .select("tweetContent", "latitude", "longitude")
      .write.csv(out + "result-" + Random.nextLong())



  }

  def analytics(spark : SparkSession) = {
    import spark.implicits._
    val out = spark.sqlContext.getConf("userOutputPath")
    val zippedDataset = spark.read.parquet(out + "zipped.parquet")
    println(zippedDataset.count())
  }

}